from zarena.zarena import BlackjackEngine  # rust module
from zarena.gym_blackjack.envs.blackjack_env import BlackjackEnv  # envs
from gym.envs.registration import register  # to register envs

register(
    id="BlackjackEnv-v1",
    entry_point="zarena.gym_blackjack:BlackjackEnv",
)



from copy import copy

import numpy as np
from zarena.gym_chess import ChessEnv
from zarena.gym_chess.envs.chess_env import (
    KING_ID,
    QUEEN_ID,
    ROOK_ID,
    BISHOP_ID,
    KNIGHT_ID,
    PAWN_ID,
)
from zarena.gym_chess.test.utils import run_test_funcs


# Blank board
BASIC_BOARD = np.array([[0] * 8] * 8, dtype=np.int8)
BASIC_BOARD[3, 3] = -PAWN_ID
BASIC_BOARD[3, 4] = -PAWN_ID
BASIC_BOARD[3, 5] = -PAWN_ID
BASIC_BOARD[6, 3] = PAWN_ID
BASIC_BOARD[6, 4] = PAWN_ID
BASIC_BOARD[6, 5] = PAWN_ID


# Pawn capture movements
def test_pawn_capture_moves():
    BOARD = copy(BASIC_BOARD)
    BOARD[4, 4] = PAWN_ID
    env = ChessEnv(initial_board=BOARD)
    moves = env.get_possible_moves()
    env.render_moves(moves)
    expected_attacks = set([(5, 5), (4, 3), (5, 4), (4, 5), (3, 3), (5, 3), (3, 5)])
    squares_attacked = set([tuple(move[1]) for move in moves])
    assert squares_attacked == expected_attacks


# Knight capture movements
def test_knight_capture_moves():
    BOARD = copy(BASIC_BOARD)
    BOARD[5, 3] = KNIGHT_ID
    env = ChessEnv(initial_board=BOARD)
    moves = env.get_possible_moves()
    env.render_moves(moves)
    expected_attacks = set(
        [
            (4, 4),
            (7, 4),
            (5, 5),
            (3, 4),
            # (4, 3),
            (6, 1),
            (5, 4),
            (4, 5),
            (7, 2),
            (3, 2),
            (4, 1),
        ]
    )
    squares_attacked = set([tuple(move[1]) for move in moves])
    print(squares_attacked)
    print(expected_attacks)
    assert squares_attacked == expected_attacks


# Bishop capture movements
def test_bishop_capture_moves():
    BOARD = copy(BASIC_BOARD)
    BOARD[5, 3] = BISHOP_ID
    env = ChessEnv(initial_board=BOARD)
    env.reset()
    moves = env.get_possible_moves()
    env.render_moves(moves)
    expected_attacks = set(
        [
            (4, 4),
            (6, 2),
            (5, 5),
            (7, 1),
            # (4, 3),
            (3, 1),
            (5, 4),
            (2, 0),
            (4, 2),
            (4, 5),
            (3, 5),
        ]
    )
    squares_attacked = set([tuple(move[1]) for move in moves])
    assert squares_attacked == expected_attacks


# Rook capture movements
def test_rook_capture_moves():
    BOARD = copy(BASIC_BOARD)
    BOARD[5, 3] = ROOK_ID
    env = ChessEnv(initial_board=BOARD)
    moves = env.get_possible_moves()
    env.render_moves(moves)
    expected_attacks = set(
        [
            (4, 4),
            (5, 5),
            (4, 3),
            (5, 4),
            (5, 1),
            (5, 7),
            (4, 5),
            (3, 3),
            (5, 0),
            (5, 6),
            (5, 2),
        ]
    )
    squares_attacked = set([tuple(move[1]) for move in moves])
    assert squares_attacked == expected_attacks


# Queen capture movements
def test_queen_capture_moves():
    BOARD = copy(BASIC_BOARD)
    BOARD[5, 3] = QUEEN_ID
    env = ChessEnv(initial_board=BOARD)
    moves = env.get_possible_moves()
    env.render_moves(moves)
    expected_attacks = set(
        [
            (4, 4),
            (6, 2),
            (5, 5),
            (7, 1),
            (4, 3),
            (3, 1),
            (5, 4),
            (2, 0),
            (5, 1),
            (5, 7),
            (4, 2),
            (4, 5),
            (3, 3),
            (5, 0),
            (5, 6),
            (3, 5),
            (5, 2),
        ]
    )
    squares_attacked = set([tuple(move[1]) for move in moves])
    assert squares_attacked == expected_attacks


# King capture movements
def test_king_capture_moves():
    BOARD = copy(BASIC_BOARD)
    BOARD[4, 3] = KING_ID
    BOARD[3, 2] = -PAWN_ID
    BOARD[2, 5] = -PAWN_ID
    BOARD[3, 5] = 0
    BOARD[6, 3] = 0
    BOARD[6, 4] = 0
    BOARD[6, 5] = 0
    env = ChessEnv(initial_board=BOARD)
    env.render()
    moves = env.get_possible_moves()
    env.render_moves(moves)
    expected_attacks = set([(3, 2), (3, 3), (5, 2), (5, 3), (5, 4)])
    squares_attacked = set([tuple(move[1]) for move in moves])
    assert squares_attacked == expected_attacks


if __name__ == "__main__":
    run_test_funcs(__name__)



import sys
import time
import random

import numpy as np
import gym
import gym_chess

env = gym.make("ChessVsSelf-v3", log=False)

#
# Play against self
#
num_episodes = 2
num_steps = 100

total_steps = 0
collected_rewards = []
start = time.time()

for i in range(num_episodes):
    initial_state = env.reset()
    print("\n", "=" * 10, "NEW GAME", "=" * 10)
    env.render()
    total_rewards = {"WHITE": 0, "BLACK": 0}

    for j in range(num_steps):
        total_steps += 1
        # white moves
        moves = env.possible_moves
        m = random.choice(moves)
        a = env.move_to_action(m)
        # perform action
        state, reward, done, _ = env.step(a)
        total_rewards["WHITE"] += reward
        if done:
            break

        # black moves
        moves = env.possible_moves
        m = random.choice(moves)
        a = env.move_to_action(m)
        # perform action
        state, reward, done, _ = env.step(a)
        total_rewards["BLACK"] += reward
        if done:
            break

    print(">" * 5, "GAME", i, "REWARD:", total_rewards)
    collected_rewards.append(total_rewards)

end = time.time()
diff = end - start

print("Total time (s)", diff)
print("Total episodes", num_episodes)
print("Total steps", total_steps)
print("Time per episode (s)", diff / num_episodes)
print("Time per step (s)", diff / total_steps)



from copy import copy

import numpy as np
from zarena.gym_chess import ChessEnv
from zarena.gym_chess.envs.chess_env import (
    KING_ID,
    QUEEN_ID,
    ROOK_ID,
    BISHOP_ID,
    KNIGHT_ID,
    PAWN_ID,
)
from zarena.gym_chess.test.utils import run_test_funcs


# Blank board
BASIC_BOARD = np.array([[0] * 8] * 8, dtype=np.int8)
BASIC_BOARD[3, 2] = -PAWN_ID
BASIC_BOARD[3, 3] = -PAWN_ID
BASIC_BOARD[3, 4] = -PAWN_ID
BASIC_BOARD[3, 5] = -PAWN_ID
BASIC_BOARD[3, 6] = -PAWN_ID
BASIC_BOARD[6, 2] = PAWN_ID
BASIC_BOARD[6, 3] = PAWN_ID
BASIC_BOARD[6, 4] = PAWN_ID
BASIC_BOARD[6, 5] = PAWN_ID
BASIC_BOARD[6, 6] = PAWN_ID


# Pawn basic movements
def test_pawn_moves():
    BOARD = copy(BASIC_BOARD)
    BOARD[4, 4] = PAWN_ID
    env = ChessEnv(initial_board=BOARD)
    moves = env.get_possible_moves(attack=True)
    env.render_moves(moves)
    expected_attacks = set(
        [(5, 5), (5, 4), (5, 1), (5, 7), (3, 3), (5, 6), (5, 3), (3, 5), (5, 2)]
    )
    squares_attacked = set([tuple(move[1]) for move in moves])
    print(squares_attacked)
    assert squares_attacked == expected_attacks


# Knight basic movements
def test_knight_moves():
    BOARD = copy(BASIC_BOARD)
    BOARD[4, 4] = KNIGHT_ID
    env = ChessEnv(initial_board=BOARD)
    moves = env.get_possible_moves(attack=True)
    env.render_moves(moves)
    expected_attacks = set(
        [
            (5, 5),
            (6, 5),
            (5, 4),
            (5, 1),
            (5, 7),
            (2, 3),
            (6, 3),
            (5, 6),
            (3, 6),
            (5, 3),
            (3, 2),
            (2, 5),
            (5, 2),
        ]
    )
    squares_attacked = set([tuple(move[1]) for move in moves])
    print(squares_attacked)
    assert squares_attacked == expected_attacks


# Bishop basic movements
def test_bishop_moves():
    BOARD = copy(BASIC_BOARD)
    BOARD[4, 4] = BISHOP_ID
    env = ChessEnv(initial_board=BOARD)
    moves = env.get_possible_moves(attack=True)
    env.render_moves(moves)
    expected_attacks = set(
        [
            (3, 3),
            (3, 5),
            (5, 3),
            (6, 2),
            (5, 5),
            (6, 6),
            (5, 3),
            (5, 1),
            (5, 4),
            (5, 2),
            (5, 5),
            (5, 3),
            (5, 6),
            (5, 4),
            (5, 7),
            (5, 5),
        ]
    )
    squares_attacked = set([tuple(move[1]) for move in moves])
    assert squares_attacked == expected_attacks


# Rook basic movements
def test_rook_moves():
    BOARD = copy(BASIC_BOARD)
    BOARD[4, 4] = ROOK_ID
    env = ChessEnv(initial_board=BOARD)
    moves = env.get_possible_moves(attack=True)
    env.render_moves(moves)
    expected_attacks = set(
        [
            (4, 0),
            (5, 5),
            (3, 4),
            (4, 3),
            (5, 4),
            (4, 6),
            (6, 4),
            (4, 2),
            (5, 1),
            (5, 7),
            (4, 5),
            (5, 6),
            (5, 3),
            (4, 1),
            (4, 7),
            (5, 2),
        ]
    )
    squares_attacked = set([tuple(move[1]) for move in moves])
    print(squares_attacked)
    assert squares_attacked == expected_attacks


# Queen basic movements
def test_queen_moves():
    BOARD = copy(BASIC_BOARD)
    BOARD[4, 4] = QUEEN_ID
    env = ChessEnv(initial_board=BOARD)
    moves = env.get_possible_moves(attack=True)
    env.render_moves(moves)
    expected_attacks = set(
        [
            (4, 0),
            (3, 4),
            (4, 3),
            (5, 4),
            (4, 6),
            (5, 1),
            (5, 7),
            (6, 2),
            (4, 2),
            (4, 5),
            (3, 3),
            (5, 6),
            (5, 3),
            (6, 4),
            (4, 1),
            (4, 7),
            (3, 5),
            (5, 2),
            (5, 5),
            (6, 6),
        ]
    )
    squares_attacked = set([tuple(move[1]) for move in moves])
    print(squares_attacked)
    assert squares_attacked == expected_attacks


# King basic movements
def test_king_moves():
    BOARD = copy(BASIC_BOARD)
    BOARD[4, 4] = KING_ID
    env = ChessEnv(initial_board=BOARD)
    moves = env.get_possible_moves(attack=True)
    env.render_moves(moves)
    expected_attacks = set(
        [
            (5, 5),
            (3, 4),
            (4, 3),
            (5, 4),
            (5, 1),
            (5, 7),
            (4, 5),
            (3, 3),
            (5, 6),
            (5, 3),
            (3, 5),
            (5, 2),
        ]
    )
    squares_attacked = set([tuple(move[1]) for move in moves])
    print(squares_attacked)
    assert squares_attacked == expected_attacks


if __name__ == "__main__":
    run_test_funcs(__name__)



from collections import defaultdict
from copy import copy

import numpy as np
from zarena.gym_chess import ChessEnv
from zarena.gym_chess.envs.chess_env import (
    KING_ID,
    ROOK_ID,
    PAWN_ID,
    CASTLE_KING_SIDE_WHITE,
    CASTLE_QUEEN_SIDE_WHITE,
)
from zarena.gym_chess.test.utils import run_test_funcs


# Blank board
BASIC_BOARD = np.array([[0] * 8] * 8, dtype=np.int8)
BASIC_BOARD[6, 0] = PAWN_ID
BASIC_BOARD[6, 1] = PAWN_ID
BASIC_BOARD[6, 2] = PAWN_ID
BASIC_BOARD[6, 3] = PAWN_ID
BASIC_BOARD[6, 4] = PAWN_ID
BASIC_BOARD[6, 5] = PAWN_ID
BASIC_BOARD[6, 6] = PAWN_ID
BASIC_BOARD[6, 7] = PAWN_ID


# King side castle
def test_king_side_castle():
    BOARD = copy(BASIC_BOARD)
    BOARD[7, 4] = KING_ID
    BOARD[7, 7] = ROOK_ID
    env = ChessEnv(initial_board=BOARD)
    moves = env.get_castle_moves(player=env.current_player)
    env.render_moves(moves)
    assert moves == [CASTLE_KING_SIDE_WHITE]


# Queen side castle
def test_queen_side_castle():
    BOARD = copy(BASIC_BOARD)
    BOARD[7, 0] = ROOK_ID
    BOARD[7, 4] = KING_ID
    env = ChessEnv(initial_board=BOARD)
    moves = env.get_castle_moves(player=env.current_player)
    env.render_moves(moves)
    assert moves == [CASTLE_QUEEN_SIDE_WHITE]


# Attacked square side castle
# def test_attacked_square_castling_path():
#     BOARD = copy(BASIC_BOARD)
#     BOARD[0, 2] = -ROOK_ID
#     BOARD[6, 2] = 0
#     BOARD[7, 0] = ROOK_ID
#     BOARD[7, 4] = KING_ID
#     env = ChessEnv(initial_board=BOARD)

#     opponent = env.get_other_player(env.current_player)
#     moves = env.get_possible_moves(player=opponent, attack=True)
#     env.render_moves(moves)

#     squares_under_attack = env.get_squares_attacked_by_player(env.state, opponent)
#     squares_under_attack_hashmap = defaultdict(lambda: None)
#     for sq in squares_under_attack:
#         squares_under_attack_hashmap[tuple(sq)] = True
#     moves = env.castle_moves(
#         env.current_player, squares_under_attack_hashmap=squares_under_attack_hashmap
#     )
#     env.render_moves(moves)
#     assert moves == []


# King moves
# def test_king_has_moved_castling():
# BOARD = copy(BASIC_BOARD)
# BOARD[7, 0] = ROOK_ID
# BOARD[7, 4] = KING_ID
# env = ChessEnv(initial_board=BOARD)
# king_moves = env.king_moves(env.current_player, np.array([7, 4]))
# action = env.move_to_action(king_moves[0])
# env.step(action)
# env.render()
# env.current_player = "white"
# moves = env.castle_moves(env.current_player)
# env.render_moves(moves)
#     assert moves == []


if __name__ == "__main__":
    run_test_funcs(__name__)



from copy import copy

import numpy as np
from zarena.gym_chess import ChessEnv
from zarena.gym_chess.envs.chess_env import (
    KING_ID,
    ROOK_ID,
    PAWN_ID,
)
from zarena.gym_chess.test.utils import run_test_funcs


# Blank board
BASIC_BOARD = np.array([[0] * 8] * 8, dtype=np.int8)
BASIC_BOARD[3, 3] = -PAWN_ID
BASIC_BOARD[2, 4] = -PAWN_ID
BASIC_BOARD[3, 5] = -PAWN_ID


# King capture movements
def test_king_moves_1():
    BOARD = copy(BASIC_BOARD)
    BOARD[4, 4] = KING_ID
    BOARD[0, 0] = ROOK_ID
    env = ChessEnv(initial_board=BOARD)
    env.reset()
    moves = env.get_possible_moves()
    env.render_moves(moves)
    env.render()
    king_is_checked = env.white_king_is_checked
    expected_attacks = set([(5, 5), (3, 4), (4, 3), (5, 4), (4, 5), (5, 3)])
    squares_attacked = set([tuple(move[1]) for move in moves])
    assert squares_attacked == expected_attacks
    assert king_is_checked


# King capture movements
def test_king_moves_2():
    BOARD = copy(BASIC_BOARD)
    BOARD[3, 4] = KING_ID
    env = ChessEnv(initial_board=BOARD)
    moves = env.get_possible_moves()
    env.render_moves(moves)
    king_is_checked = env.white_king_is_checked
    expected_attacks = set([(2, 4), (4, 3), (2, 3), (4, 5), (2, 5)])
    squares_attacked = set([tuple(move[1]) for move in moves])
    assert squares_attacked == expected_attacks
    assert not king_is_checked


if __name__ == "__main__":
    run_test_funcs(__name__)



from zarena.zarena import ChessEngine  # rust module
from zarena.gym_chess.envs.chess_env import ChessEnv  # envs
from gym.envs.registration import register  # to register envs

register(id="ChessEnv-v3", entry_point="zarena.gym_chess:ChessEnv")



import inspect
import sys


def run_test_funcs(namespace):
    test_funcs = [
        obj
        for name, obj in inspect.getmembers(sys.modules[namespace])
        if (inspect.isfunction(obj) and name.startswith("test_"))
    ]
    for func in test_funcs:
        func()



from zarena.zarena import TictactoeEngine  # rust module
from zarena.gym_tictactoe.envs.tictactoe import TictactoeEnv  # envs
from gym.envs.registration import register  # to register envs

register(
    id="GatoEnv-v2",
    entry_point="zarena.gym_tictactoe:TictactoeEnv",
)



import gym
from zarena.gym_poker import PokerEngine

# import random
# from tqdm import tqdm

# POKER GYM ENVIRONMENT CLASS
# ---------------------------
class PokerEnv(gym.Env):
    """
    Game wrapper.
    """

    def __init__(self, n_players=2, infinite_game=True):
        # engine
        self.engine = PokerEngine(n_players, infinite_game)

    def step(self, action):
        """
        Apply action to the game.

        Args:
            action : action of the action_space to take.

        Returns:
            The new observation, the reward and a boolean if the game has ended.
        """
        observation, reward, done = self.engine.step(action)
        return observation, reward, done, None

    def to_play(self):
        """
        Return the current player.

        Returns:
            The current player, it should be an element of the players list in the config.
        """
        return self.engine.to_play()

    def legal_actions(self):
        """
        Should return the legal actions at each turn, if it is not available, it can return
        the whole action space. At each turn, the game have to be able to handle one of returned actions.

        For complex game where calculating legal moves is too long, the idea is to define the legal actions
        equal to the action space but to return a negative reward if the action is illegal.

        Returns:
            An array of integers, subset of the action space.
        """
        return self.engine.legal_actions()

    def reset(self):
        """
        Reset the game for a new game.

        Returns:
            Initial observation of the game.
        """
        return self.engine.reset()

    def get_state(self):
        """
        Get the current state of the game
        """
        state_a, state_b = self.engine.get_state_a(), self.engine.get_state_b()
        community_cards = []
        if len(state_a[0]) > 0:
            community_cards = self.hand_to_string(state_a[0])
        players_hand = []
        for player in state_a[1]:
            players_hand.append(self.hand_to_string(player["hand"]["cards"]))
        lista_a = list(state_a)
        lista_a[0] = community_cards
        for player in lista_a[1]:
            player["hand"] = players_hand[player["id"]]
        state_a = tuple(lista_a)
        state = state_a + state_b
        return state

    def get_total_players(self):
        """
        Returns:
            Integer total players + dealer
        """
        return self.engine.get_total_players()

    def card_to_string(self, card):
        figure = card // 15
        value = card % 15
        if value == 11:
            v_str = "J"
        elif value == 12:
            v_str = "Q"
        elif value == 13:
            v_str = "K"
        elif value == 14:
            v_str = "A"
        else:
            v_str = str(value)

        if figure == 0:
            f_str = "-♥"
        elif figure == 1:
            f_str = "-♠"
        elif figure == 2:
            f_str = "-♣"
        elif figure == 3:
            f_str = "-♦"
        else:
            f_str = "was over"

        return v_str + f_str

    def hand_to_string(self, hand):
        res = []
        for card in hand:
            res.append(self.card_to_string(card))
        return res



import os
import sys
import datetime
import numpy
import toml
import gym

from collections import defaultdict
from copy import copy
from dataclasses import dataclass
from six import StringIO
from pprint import pprint
from gym import spaces, error, utils
from gym.utils import seeding

from zarena.gym_chess import ChessEngine

EMPTY_SQUARE_ID = 0
KING_ID = 1
QUEEN_ID = 2
ROOK_ID = 3
BISHOP_ID = 4
KNIGHT_ID = 5
PAWN_ID = 6

KING = "king"
QUEEN = "queen"
ROOK = "rook"
BISHOP = "bishop"
KNIGHT = "knight"
PAWN = "pawn"

KING_DESC = "K"
QUEEN_DESC = "Q"
ROOK_DESC = "R"
BISHOP_DESC = "B"
KNIGHT_DESC = "N"
PAWN_DESC = ""

WHITE_ID = 1
BLACK_ID = -1

WHITE = "WHITE"
BLACK = "BLACK"

WIN_REWARD = 1
LOSS_REWARD = 0
DRAW_REWARD = 0.5
INVALID_ACTION_REWARD = -1


@dataclass
class Piece:
    id: int
    icon: str
    desc: str
    type: str
    color: str


PIECES = [
    Piece(icon="♙", desc=PAWN_DESC, color=BLACK, type=PAWN, id=-PAWN_ID),
    Piece(icon="♘", desc=KNIGHT_DESC, color=BLACK, type=KNIGHT, id=-KNIGHT_ID),
    Piece(icon="♗", desc=BISHOP_DESC, color=BLACK, type=BISHOP, id=-BISHOP_ID),
    Piece(icon="♖", desc=ROOK_DESC, color=BLACK, type=ROOK, id=-ROOK_ID),
    Piece(icon="♕", desc=QUEEN_DESC, color=BLACK, type=QUEEN, id=-QUEEN_ID),
    Piece(icon="♔", desc=KING_DESC, color=BLACK, type=KING, id=-KING_ID),
    Piece(icon=".", desc="", color=None, type=None, id=EMPTY_SQUARE_ID),
    Piece(icon="♚", desc=KING_DESC, color=WHITE, type=KING, id=KING_ID),
    Piece(icon="♛", desc=QUEEN_DESC, color=WHITE, type=QUEEN, id=QUEEN_ID),
    Piece(icon="♜", desc=ROOK_DESC, color=WHITE, type=ROOK, id=ROOK_ID),
    Piece(icon="♝", desc=BISHOP_DESC, color=WHITE, type=BISHOP, id=BISHOP_ID),
    Piece(icon="♞", desc=KNIGHT_DESC, color=WHITE, type=KNIGHT, id=KNIGHT_ID),
    Piece(icon="♟", desc=PAWN_DESC, color=WHITE, type=PAWN, id=PAWN_ID),
]

ID_TO_COLOR = {piece.id: piece.color for piece in PIECES}
ID_TO_ICON = {piece.id: piece.icon for piece in PIECES}
ID_TO_TYPE = {piece.id: piece.type for piece in PIECES}
ID_TO_DESC = {piece.id: piece.desc for piece in PIECES}

RESIGN = "RESIGN"
CASTLE_KING_SIDE_WHITE = "CASTLE_KING_SIDE_WHITE"
CASTLE_QUEEN_SIDE_WHITE = "CASTLE_QUEEN_SIDE_WHITE"
CASTLE_KING_SIDE_BLACK = "CASTLE_KING_SIDE_BLACK"
CASTLE_QUEEN_SIDE_BLACK = "CASTLE_QUEEN_SIDE_BLACK"
CASTLE_MOVES = [
    CASTLE_KING_SIDE_WHITE,
    CASTLE_QUEEN_SIDE_WHITE,
    CASTLE_KING_SIDE_BLACK,
    CASTLE_QUEEN_SIDE_BLACK,
]

DEFAULT_BOARD = [
    [-3, -5, -4, -2, -1, -4, -5, -3],
    [-6, -6, -6, -6, -6, -6, -6, -6],
    [0] * 8,
    [0] * 8,
    [0] * 8,
    [0] * 8,
    [6, 6, 6, 6, 6, 6, 6, 6],
    [3, 5, 4, 2, 1, 4, 5, 3],
]

FILE_NAMES = ["a", "b", "c", "d", "e", "f", "g", "h"]
RANK_NAMES = ["1", "2", "3", "4", "5", "6", "7", "8"]


def highlight(string, background="white", color="gray"):
    return utils.colorize(utils.colorize(string, color), background, highlight=True)


# CHESS GYM ENVIRONMENT CLASS
# ---------------------------
class ChessEnv(gym.Env):
    def __init__(self, player_color=WHITE, log=True, initial_board=DEFAULT_BOARD):

        # constants
        self.log = log
        self.initial_board = initial_board
        self.board = self.initial_board

        # engine
        self.engine = ChessEngine()

        # Muzero control of players
        self.player = 1
        self.player_color = player_color  # define player # TODO: implement
        self.current_player = player_color
        self.white_king_castle_is_possible = True
        self.white_queen_castle_is_possible = True
        self.black_king_castle_is_possible = True
        self.black_queen_castle_is_possible = True
        self.white_king_is_checked = False
        self.black_king_is_checked = False

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)

        return [seed]

    def to_play(self):
        return 0 if self.player == 1 else 1

    def reset(self):
        """
        Resets the state of the environment, returning an initial observation.
        Outputs -> observation : the initial observation of the space. (Initial reward is assumed to be 0.)
        """
        self.board = self.initial_board
        # self.prev_board = None ## TODO: use prev_board to check for en-passant mvoes
        self.done = False
        self.player = 1
        self.current_player = WHITE
        self.saved_boards = defaultdict(lambda: 0)
        self.repetitions = 0  # 3 repetitions ==> DRAW
        self.states = []
        self.move_count = 0
        self.white_king_castle_is_possible = True
        self.white_queen_castle_is_possible = True
        self.black_king_castle_is_possible = True
        self.black_queen_castle_is_possible = True
        self.white_king_is_checked = False
        self.black_king_is_checked = False
        self.white_king_on_the_board = self.piece_is_on_board(self.board, KING_ID)
        self.black_king_on_the_board = self.piece_is_on_board(self.board, -KING_ID)
        # update state with engine
        self.state = self.engine.update_state(self.state)
        # pre-calculate possible moves
        self.possible_moves = self.get_possible_moves(state=self.state, player=WHITE)
        # If player chooses black, make white opponnent move first
        if self.player == BLACK:
            white_first_move = self.opponent_policy(self)
            white_first_action = self.move_to_action(white_first_move)
            # make move
            # self.board, _, _, _ = self.step(white_first_action)
            self.state, _, _ = self.player_move(white_first_action)
            self.move_count += 1
            self.current_player = BLACK
            self.possible_moves = self.get_possible_moves(
                state=self.state, player=BLACK
            )

        return self.get_observation()

    def step(self, action):
        """
        Run one timestep of the environment's dynamics. When end of episode
        is reached, reset() should be called to reset the environment's internal state.

        Input
        -----
        action : an action provided by the environment

        Outputs
        -------
        (observation, reward, done, info)
        observation : agent's observation of the current environment
        reward [Float] : amount of reward due to the previous action
        done : a boolean, indicating whether the episode has ended
        info : a dictionary containing other diagnostic information from the previous action
        """

        # action invalid in current state
        self.possible_moves = self.get_possible_moves()
        if action not in self.possible_actions:
            return self.get_observation, INVALID_ACTION_REWARD, self.done, self.info

        # Game is done
        if self.done:
            return self.get_observation, DRAW_REWARD, True, self.info

        # make move
        self.state, reward = self.player_move(action)

        # 3-fold repetition => DRAW
        self.states.append(self.state)
        if self.engine.is_game_over(self.states, self.state, self.current_player) != 0:
            self.done = True

        if self.have_winner() or len(self.legal_actions(self.switch_player())) == 0:
            self.done = True
        reward = self.get_reward() if self.done else 0

        if self.done:
            return self.get_observation(), reward, True, self.info

        # Change Player Color
        self.current_player = self.switch_player()

        # Change player Number
        self.player *= -1

        if self.current_player == WHITE:
            self.move_count += 1

            print("# of turns: " + str(self.move_count))

        return self.get_observation(), reward, self.done, self.info

    def have_winner(self):

        # check if there are no possible_moves for self
        # AKA lose
        if not self.possible_moves and self.king_is_checked(player=self.current_player):
            return True

        # AKA WIN
        opponent_player = self.switch_player()
        self.possible_moves = self.get_possible_moves(player=opponent_player)
        # check if there are no possible_moves for opponent
        if not self.possible_moves and self.king_is_checked(player=opponent_player):
            return True

        return False

    def get_reward(self):
        # check if there are no possible_moves for self
        # LOSE
        # if not self.possible_moves and self.king_is_checked(player=self.current_player):
        #     return LOSS_REWARD

        # WIN
        # opponent_player = self.switch_player()
        # self.possible_moves = self.get_possible_moves(player=opponent_player)
        # check if there are no possible_moves for opponent
        # if not self.possible_moves and self.king_is_checked(player=opponent_player):
        #     return WIN_REWARD
        player = self.current_player
        other_player = self.switch_player()
        state = self.state
        states = self.states
        # call Rust binary
        if self.engine.is_game_over(states, state, player) == 1:
            return LOSS_REWARD
        if self.engine.is_game_over(states, state, other_player) == 1:
            return WIN_REWARD
        return DRAW_REWARD

    def switch_player(self):
        other_player = self.get_other_player(self.current_player)
        return other_player

    @property
    def state(self):
        return dict(
            board=self.board,
            # prev_board=self.prev_board,
            current_player=self.current_player,
            white_king_castle_is_possible=self.white_king_castle_is_possible,
            white_queen_castle_is_possible=self.white_queen_castle_is_possible,
            black_king_castle_is_possible=self.black_king_castle_is_possible,
            black_queen_castle_is_possible=self.black_queen_castle_is_possible,
            white_king_is_checked=self.white_king_is_checked,
            black_king_is_checked=self.black_king_is_checked,
        )

    @state.setter
    def state(self, state):
        self.board = state.get("board")
        self.white_king_castle_is_possible = state.get("white_king_castle_is_possible")
        self.white_queen_castle_is_possible = state.get(
            "white_queen_castle_is_possible"
        )
        self.black_king_castle_is_possible = state.get("black_king_castle_is_possible")
        self.black_queen_castle_is_possible = state.get(
            "black_queen_castle_is_possible"
        )
        self.white_king_is_checked = state.get("white_king_is_checked")
        self.black_king_is_checked = state.get("black_king_is_checked")

    @property
    def possible_moves(self):
        return self._possible_moves

    @possible_moves.setter
    def possible_moves(self, moves):
        self._possible_moves = moves

    @property
    def possible_actions(self):
        return [self.move_to_action(m) for m in self.possible_moves]

    @property
    def info(self):
        return dict(
            # **self.state,
            move_count=self.move_count,
            current_player=self.current_player,
            possible_moves=self.possible_moves,
            # possible_actions=self.possible_actions,
            white_king_castle_is_possible=self.white_king_castle_is_possible,
            white_queen_castle_is_possible=self.white_queen_castle_is_possible,
            black_king_castle_is_possible=self.black_king_castle_is_possible,
            black_queen_castle_is_possible=self.black_queen_castle_is_possible,
            white_king_is_checked=self.white_king_is_checked,
            black_king_is_checked=self.black_king_is_checked,
            white_king_on_the_board=self.white_king_on_the_board,
            black_king_on_the_board=self.black_king_on_the_board,
        )

    @property
    def opponent_player(self):
        if self.current_player == WHITE:
            return BLACK
        return WHITE

    @property
    def current_player_is_white(self):
        return self.current_player == WHITE

    @property
    def current_player_is_black(self):
        return not self.current_player_is_white

    def king_is_checked(self, player):
        if player == WHITE:
            return self.white_king_is_checked
        else:
            return self.black_king_is_checked

    def is_checkmate(self):
        if not self.king_is_checked():
            return False

    def piece_is_on_board(self, board, piece_id):
        for row in board:
            for square in row:
                if square == piece_id:
                    return True
        return False

    def player_can_castle(self, player):
        if player == WHITE:
            return (
                self.white_king_castle_is_possible
                and self.white_queen_castle_is_possible
            )
        else:
            return (
                self.black_king_castle_is_possible
                and self.black_queen_castle_is_possible
            )

    def get_other_player(self, player):
        if player == WHITE:
            return BLACK
        return WHITE

    def player_move(self, action):
        """
        Returns (state, reward, done)
        """
        # Resign
        if self.is_resignation(action):
            return self.state, LOSS_REWARD
        # Play
        move = self.action_to_move(action)
        new_state = self.next_state(self.state, self.current_player, move)
        # Render
        if self.log:
            print(" " * 10, ">" * 10, self.current_player)
            self.render_moves([move], mode="human")
        return new_state, LOSS_REWARD

    def next_state(self, state, player, move):
        if state is None:
            state = self.state
        move_str = self.move_to_str_code(move)
        # breakpoint()
        next_state = self.engine.next_state(state, player, move_str)
        return next_state

    def board_to_grid(self):
        grid = [[f" {ID_TO_ICON[square]} " for square in row] for row in self.board]
        return grid

    def render_grid(self, grid, mode="human"):
        outfile = sys.stdout if mode == "human" else StringIO()
        outfile.write("    ")
        outfile.write("-" * 25)
        outfile.write("\n")
        rows = "87654321"
        for i, row in enumerate(grid):
            outfile.write(f" {rows[i]} | ")
            for square in row:
                outfile.write(square)
            outfile.write("|\n")
        outfile.write("    ")
        outfile.write("-" * 25)
        outfile.write("\n      a  b  c  d  e  f  g  h ")
        outfile.write("\n")

        if mode == "string":
            return outfile.getvalue()
        if mode != "human":
            return outfile

    def render(self, mode="human"):
        """Render the playing board"""
        grid = self.board_to_grid()
        out = self.render_grid(grid, mode=mode)
        return out

    def render_moves(self, moves, mode="human"):
        grid = self.board_to_grid()
        for move in moves:
            if type(move) is str and move in CASTLE_MOVES:
                if move == CASTLE_QUEEN_SIDE_WHITE:
                    grid[7][0] = highlight(grid[7][0], background="white")
                    grid[7][1] = highlight(" >>", background="green")
                    grid[7][2] = highlight("> <", background="green")
                    grid[7][3] = highlight("<< ", background="green")
                    grid[7][4] = highlight(grid[7][4], background="white")
                elif move == CASTLE_KING_SIDE_WHITE:
                    grid[7][4] = highlight(grid[7][4], background="white")
                    grid[7][5] = highlight(" >>", background="green")
                    grid[7][6] = highlight("<< ", background="green")
                    grid[7][7] = highlight(grid[7][7], background="white")
                elif move == CASTLE_QUEEN_SIDE_BLACK:
                    grid[0][0] = highlight(grid[0][0], background="white")
                    grid[0][1] = highlight(" >>", background="green")
                    grid[0][2] = highlight("> <", background="green")
                    grid[0][3] = highlight("<< ", background="green")
                    grid[0][4] = highlight(grid[0][4], background="white")
                elif move == CASTLE_KING_SIDE_BLACK:
                    grid[0][4] = highlight(grid[0][4], background="white")
                    grid[0][5] = highlight(" >>", background="green")
                    grid[0][6] = highlight("<< ", background="green")
                    grid[0][7] = highlight(grid[0][7], background="white")
                continue

            x0, y0 = move[0][0], move[0][1]
            x1, y1 = move[1][0], move[1][1]
            if len(grid[x0][y0]) < 4:
                grid[x0][y0] = highlight(grid[x0][y0], background="white")
            if len(grid[x1][y1]) < 4:
                if self.board[x1][y1]:
                    grid[x1][y1] = highlight(grid[x1][y1], background="red")
                else:
                    grid[x1][y1] = highlight(grid[x1][y1], background="green")
        return self.render_grid(grid, mode=mode)

    def move_to_action(self, move):
        if type(move) in [list, tuple]:
            _from = move[0][0] * 8 + move[0][1]
            _to = move[1][0] * 8 + move[1][1]
            return _from * 64 + _to
        if move == CASTLE_KING_SIDE_WHITE:
            return 64 * 64
        elif move == CASTLE_QUEEN_SIDE_WHITE:
            return 64 * 64 + 1
        elif move == CASTLE_KING_SIDE_BLACK:
            return 64 * 64 + 2
        elif move == CASTLE_QUEEN_SIDE_BLACK:
            return 64 * 64 + 3
        elif move == RESIGN:
            return 64 * 64 + 4

    def action_to_move(self, action):
        return self._action_to_move(action, as_string=False)

    def action_to_move_str(self, action):
        return self._action_to_move(action, as_string=True)

    def _action_to_move(self, action, as_string=False):
        if action >= 64 * 64:
            _action = action - 64 * 64
            if _action == 0:
                return CASTLE_KING_SIDE_WHITE
            elif _action == 1:
                return CASTLE_QUEEN_SIDE_WHITE
            elif _action == 2:
                return CASTLE_KING_SIDE_BLACK
            elif _action == 3:
                return CASTLE_QUEEN_SIDE_BLACK
            elif _action == 4:
                return RESIGN
        _from, _to = action // 64, action % 64
        x0, y0 = _from // 8, _from % 8
        x1, y1 = _to // 8, _to % 8
        if not as_string:
            return ((x0, y0), (x1, y1))
        return self.move_to_str_code(action)

    def move_to_str_code(self, move):
        if move in CASTLE_MOVES:
            return move
        (x0, y0), (x1, y1) = move
        rows = list(reversed("12345678"))
        cols = "abcdefgh"
        return f"{cols[y0]}{rows[x0]}{cols[y1]}{rows[x1]}"

    def move_to_string(self, move):
        if move in [CASTLE_KING_SIDE_WHITE, CASTLE_KING_SIDE_BLACK]:
            return "O-O"
        elif move in [CASTLE_QUEEN_SIDE_WHITE, CASTLE_QUEEN_SIDE_BLACK]:
            return "O-O-O"
        _from, _to = move
        rows = list(reversed("12345678"))
        cols = "abcdefgh"
        piece_id = self.board[_from[0]][_from[1]]
        piece_desc = ID_TO_DESC[piece_id]
        capture = self.board[_to[0]][_to[1]] != 0
        _from_str = cols[_from[1]] + rows[_from[0]]
        _to_str = cols[_to[1]] + rows[_to[0]]
        string = f"{piece_desc}{_from_str}{'x' if capture else ''}{_to_str}"
        return string

    def rust_move_to_coords(self, move):
        if move in CASTLE_MOVES:
            return move
        cols = dict(a=0, b=1, c=2, d=3, e=4, f=5, g=6, h=7)
        _from = move[:2]
        _to = move[2:]
        _from = (8 - int(_from[1]), cols[_from[0]])
        _to = (8 - int(_to[1]), cols[_to[0]])
        move = (_from, _to)
        return move

    def get_possible_actions(self, player):
        moves = self.get_possible_moves(None, player)
        return [self.move_to_action(move) for move in moves]

    def get_possible_moves(self, state=None, player=None, attack=False):
        if state is None:
            state = self.state
        if player is None:
            player = self.current_player
        # call Rust binary
        moves = self.engine.get_possible_moves(state, player, attack)
        # print(moves)
        moves = [self.rust_move_to_coords(move) for move in moves]
        return moves

    def get_castle_moves(self, state=None, player=None):
        if state is None:
            state = self.state
        if player is None:
            player = self.current_player
        # call Rust binary
        moves = self.engine.get_castle_moves(state, player)
        # print(moves)
        moves = [self.rust_move_to_coords(move) for move in moves]
        return moves

    # TODO: implement resignation action parsing
    def is_resignation(self, action):
        return False

    @staticmethod
    def player_to_int(player):
        if player == WHITE:
            return 1
        return -1

    def encode_board(self):
        mapping = "0ABCDEFfedcba"
        encoding = "".join([mapping[val] for row in self.board for val in row])
        return encoding

    def get_observation(self):
        # TODO - Instead of entire state, player1 and player2
        # board_player1=numpy.where(self.board == White)
        # board_player2=numpy.where(self.board == Black)
        return [
            # TODO - Only Player 1
            self.board,
            # Only to Player 2
            self.board,
        ]

    def legal_actions(self, player=None):
        if player == None:
            player = self.current_player
        # Not all possible actions are legal but good enough for now
        return self.get_possible_actions(player)

    def close(self):
        pass


# import random
# game = ChessEnvV3()
# reward = []
# for i in range(0, 10):
#     done = False
#     game.reset()
#     reward_3 = 0
#     while not done:
#         # print(game.legal_actions())
#         action = random.choice(game.legal_actions())
#         _, reward_2, done, _ = game.step(action)
#         reward_3 = reward_2
#     reward.append(reward_2)
#     print(f"Reward: {reward}")



import time
from copy import copy

import numpy as np
from zarena.gym_chess import ChessEnv
from zarena.gym_chess.test.utils import run_test_funcs


def test_benchmark():
    env = ChessEnv(log=False)

    num_episodes = 10
    num_steps = 100
    total_steps = 0
    start = time.time()

    for e in range(num_episodes):
        env.reset()
        for step in range(num_steps):
            total_steps += 1

            moves = env.possible_moves
            if not moves:
                break

            idx = np.random.choice(np.arange(len(moves)))
            move = moves[idx]
            action = env.move_to_action(move)
            state, reward, done, info = env.step(action)
            if done:
                break

    end = time.time()
    diff = end - start

    print("Total time (s)", diff)
    print("Total episodes", num_episodes)
    print("Total steps", total_steps)
    print("Time per episode (s)", diff / num_episodes)
    print("Time per step (s)", diff / total_steps)

    # assert that it's less than 50 seconds
    assert diff < 10


# Total time (s) 0.5776240825653076
# Total episodes 10
# Total steps 1851
# Time per episode (s) 0.05776240825653076
# Time per step (s) 0.00031206055243938825


if __name__ == "__main__":
    run_test_funcs(__name__)



from copy import copy

import numpy as np
from zarena.gym_chess import ChessEnv
from zarena.gym_chess.envs.chess_env import PAWN_ID, DEFAULT_BOARD
from zarena.gym_chess.test.utils import run_test_funcs


# Blank board
BASIC_BOARD = np.array(DEFAULT_BOARD)

# Pawn basic movements
def test_pawn_basic_moves():
    BOARD = copy(BASIC_BOARD)
    BOARD[6, 0] = PAWN_ID
    BOARD[1, 0] = -PAWN_ID
    env = ChessEnv(initial_board=BOARD)
    env.reset()
    # player_1
    actions = env.legal_actions()
    env.step(actions[0])
    # player_2
    actions = env.legal_actions()
    env.step(actions[-1])
    # player_3
    actions = env.legal_actions()
    env.step(actions[0])
    # player_4
    actions = env.legal_actions()
    env.step(actions[-1])
    env.render()

    EXPECTED_BOARD = copy(BASIC_BOARD)
    EXPECTED_BOARD[6, 0] = 0
    EXPECTED_BOARD[4, 0] = PAWN_ID
    EXPECTED_BOARD[1, 7] = 0
    EXPECTED_BOARD[4, 7] = -PAWN_ID
    print(np.array(EXPECTED_BOARD))
    assert (np.array(env.state["board"]) == np.array(EXPECTED_BOARD)).all()


if __name__ == "__main__":
    run_test_funcs(__name__)



from zarena.zarena import PokerEngine  # rust module
from zarena.gym_poker.envs.poker_env import PokerEnv  # envs
from gym.envs.registration import register  # to register envs

register(
    id="PokerEnv-v1",
    entry_point="zarena.gym_poker:PokerEnv",
)
